% function [bufs, samples, iter]  = opt_buffer6(b,distrib)
% MATLAB statistical toolbox is required to run this routine
% The function call wip.m 
% test data (machines distribution) is available in DIST_MACHINE.mat
% b - number of buffer spaces to allocate
% distrib = array with {{'distribution name m1', mean_m1, variance_m2}, ... } as in the parameters of 
%           the function random  {{'norm',6,2},...} by defualt.
%           A sample of the structure vectore <distrib> is located in the file DIST_MACHINE.mat
%
% bufs - best buffer allocation found
% samples - number of simulations preformed
%


function [bufs, samples, iter] = opt_buffer6(b,distrib)

m = length(distrib);            %  number of machines in shop
p = ones(b+1,m-1) / (b+1);      % probability matrix (p_ij) is the probability to allocate i buffer spaces in nich j
N = ceil(5*sqrt(m*b));			% number of sample (simulations) in each iteration
alfa = 0.7;                     % smoothing factor  
gamma_ind = N - ceil(N/10);     % The top results start at gamma_ind 
sample_size = 700;				% number of jobs in each simulation
warm_up = 100;					% number of first jobs to ignore in performance measure calculation
iter = 0;
samples = 0;

while 1
   
   iter = iter + 1;
   
   data = zeros(N,m);
   
   % randomize processing times according to the given distribution. 
   % The processing time remians constant for the whole iteration
   time_table = create_time_table_for_wip(distrib,sample_size+b+m+1);
   p
   
   for i = 1:N
      
      % build a realization of buffer allocation
      p_adj = p;
      b_left = b;
      niche_left = m-1;
      niche_done = zeros(1,m-1);
      bufs = zeros(1,m-1);
      while (b_left > 0) & (niche_left > 1)
         buf_ind = find(niche_done == 0);
         niche_now = buf_ind(round(.5+rand*(niche_left)));
         niche_done(niche_now) = 1;
         p_adj_cum = p_adj;
         for ii = 2:(b_left+1)
            p_adj_cum(ii,:) = p_adj_cum(ii,:) + p_adj_cum(ii-1,:);
         end
         bufs(niche_now) = min(find(p_adj_cum(:,niche_now) >=  rand))-1;
         b_left = b_left - bufs(niche_now);
         p_adj = p_adj(1:b_left+1,:);
         niche_left = niche_left - 1;
         for ii = 1:(m-1)
            p_adj(:,ii) = p_adj(:,ii) / sum(p_adj(:,ii));
         end
      end
      
      if b_left > 0
         niche_now = find(niche_done == 0);
         bufs(niche_now) = b_left;
      end
      
      
      % run sample simulation and collect data
      data(i, 2:m) = bufs ;
      ez = find(all((data(1:(i-1),2:m) == bufs(ones(i-1,1),:))'));  
      % check wheter this buffer configurtion was already sampled to save CPU time  
      if length(ez) == 0
         data(i,1) = wip(bufs,0,time_table,warm_up,sample_size);
         samples = samples + 1;
      else
         data(i,1) = data(ez(1),1);
         fprintf(':-) ');
      end
      fprintf('%3.0f\\%3.0f  ', iter, i);
      fprintf('throughput: %2.7f  ', data(i,1));
      fprintf('buffers:');
      fprintf('%3.0f  ',data(i,2:m));
      fprintf('\n');
   end
   
   data = sortrows(data);	
   data(gamma_ind:N,:)
   
   old_p = p;
   p = zeros(b+1,m-1);
   % the best resultes are weighted by their objective function squer value
   for j = gamma_ind:N
      for i = 2:m
         p(data(j,i)+1,i-1) = p(data(j,i)+1,i-1)  + data(j,1)^2;
      end
   end
   
   % normalize the probability matrix
   ez = sum(p);
   for j = 1:(m-1)
      p(:,j) = p(:,j) ./ ez(j);
   end
   
   p = alfa*p + (1-alfa)*old_p; % weighting old probability matrix with the new one.
   
   if min(max(p)) >= 0.9
      break
   end
end

p
[ez ind] = max(p);
bufs = ind - 1;
