% simulate the effect of Work In Proccess Location on the trouput of a m machines flowshop
%
% [rate, state_struc] = WIP(buf_size, lEcho,time_table, warm_up, sample_size, state_struc)
%
% buf_size - row vector of m-1 elements containing the buffers size
%
% lEcho - boolean paramter. If true the program print the trouhput of avery 25 jobs outputed
% 			 this helps to determine the warm period needed (see bellow) 
%
% time_table - array witm column for each machine. each row contains the processing time of 
%              all the operatins to be caried on this machime
%
% warm_up -     number of finished job to ignore at the start of the simulation.  The first few tens of
%               jobs, before the system stabilize must be ignore. 50-100 is found to be generaly suffice.
%
% sample_size - number of job to complete (including warm up job). 
%
% state_struc  - a structure describe the state of the system in terms of buffers contence, remianining 
%                proc. times and timer. can be used both as input and output in cases that a 
%                certein simulation should be preformed in few unconitnues phases.
% rate - the average number of jobs output in a unit of time (after warm_up)


function [rate, state_struc] = WIP(buf_size, lEcho,time_table, warm_up, sample_size, state_struc)

buf_size = [inf , buf_size, inf];
m = length(buf_size)-1;
if (nargin < 6)
   state_struc.timer = 0;
end

if state_struc.timer == 0
   timer = 0;
   time25 =timer;
   machine_counter = zeros(m,1);
   job_count = 0;
   buf_now = zeros(m+1,1); % the acctual number of jobs waiting in each buffer
   buf_now(1) = inf;
   machine_stat = zeros(m,1); % the state of each machine 0=Ready, 1=Work, 2=Blocked 
   machine_due = ones(m,1)*inf;  % time when the each working machine is due to finish the current job
   save_timer = 0;
else
   timer = state_struc.timer;
   machine_counter = state_struc.machine_counter;
   job_count = state_struc.job_count;
   buf_now = state_struc.buf_now;
   machine_stat = state_struc.machine_stat;
   machine_due = state_struc.machine_due;
   save_timer = state_struc.save_timer;
   time25 = state_struc.time25;
end

if nargin < 4
   warm_up = 50;
end

if nargin < 5
   sample_size = 600;
end

while 1
   for i = [m:-1:1]
      
      % Send job to buffer
      if (machine_stat(i) == 2)  &  (buf_now(i+1) < buf_size(i+1))
         buf_now(i+1) = buf_now(i+1) + 1;
         machine_stat(i) = 0;
         machine_due(i) = inf;
      end
      
      % Get job from buffer
      if (machine_stat(i) == 0) &  (buf_now(i) >0)
         buf_now(i) = buf_now(i) - 1;
         machine_stat(i) = 1;
         machine_counter(i) = machine_counter(i) + 1;
         machine_due(i) = timer + time_table(machine_counter(i),i);
      end
      
      % Get job from prev machine
      if (i > 1) & (machine_stat(i) == 0) &  (machine_stat(i-1) == 2)
         machine_stat(i) = 1;
         machine_counter(i) = machine_counter(i) + 1;
         machine_due(i) = timer + time_table(machine_counter(i),i);
         machine_stat(i-1) = 0;
         machine_due(i-1) = inf;
      end
   end
   [timer i] = min(machine_due);
   machine_stat(i) = 2;
   machine_due(i) = inf;
   
   if i == m
      job_count = job_count + 1;
      if lEcho & (mod(job_count, 25) == 0)
         [timer - time25, job_count, 25/(timer-time25)]
         time25 = timer;
      end
      if job_count >= sample_size
         break;
      end
      if job_count == warm_up
         save_timer = timer;
      end
   end
   
end

rate = (sample_size - warm_up) / (timer -save_timer);

if nargin >= 2
   state_struc.timer 			 = timer;
   state_struc.machine_counter = machine_counter;
   state_struc.job_count 	    = job_count;
   state_struc.buf_now  		 = buf_now;
   state_struc.machine_stat 	 = machine_stat;
   state_struc.machine_due 	 = machine_due;
   state_struc.save_timer 		 = save_timer;
   state_struc.time25 			 = time25;
end